#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtk.h>

#include <glib.h>
#if GLIB_MAJOR_VERSION >= 2 && GLIB_MINOR_VERSION >= 6
#include <glib/gstdio.h>
#else
#include "gstdio.h"
#include "gkeyfile.h"
#endif

#include "radar.h"

/*
 * Save plot as image:
 *
 * Plot into GdkPixmap, convert to GdkPixbuf, save as any Format Image.
 */
static void
translate_point(double x1, double y1, double cx1, double cy1, int r1,
		double cx2, double cy2, int r2, double *x2, double *y2)
{
	*x2 = cx2 + (x1 - cx1) * ((double) r2) / ((double) r1);
	*y2 = cy2 + (y1 - cy1) * ((double) r2) / ((double) r1);
}

static void
translate_length(double l1, int r1, int r2, double *l2)
{
	*l2 = l1 * ((double) r2) / ((double) r1);
}

int
radar_save_as_image(radar_t *radar, int width, int height,
		    const char *filename, const char *type,
		    char **keys, char **values)
{
	PangoFontDescription *font_description;
	PangoContext *context;
	PangoAttrList *attrs;
	PangoLayout *layout;
	char label_font[32];
	GdkPixbuf *pixbuf;
	GError *gerror;
	sights_t *s;
	vector_t *v, tvec;
	arc_t *a, tarc;
	poly_t *p, tpoly;
	char text[16];
	double cx, cy;
	int step, radius;
	int tw, th;
	int base, size;
	int error = 0;
	int i, j, k;

	if (width < height)
		base = width;
	else
		base = height;

	size = (base / 60);
// printf("%s: base %u, size %u\n", __FUNCTION__, base, size);
	if (size < 8)
		size = 8;
	if (size > 24)
		size = 24;
// printf("%s: base %u, size %u\n", __FUNCTION__, base, size);

	sprintf(label_font, "%s %u", LABEL_BASE, size);

	context = gtk_widget_get_pango_context(radar->canvas);
	layout = pango_layout_new(context);
	font_description = pango_font_description_from_string(label_font);

	pango_layout_set_alignment(layout, PANGO_ALIGN_CENTER);
	pango_layout_set_font_description(layout, font_description);

	attrs = pango_attr_list_new();
	pango_attr_list_insert(attrs, pango_attr_foreground_new(0, 0, 0));
	pango_attr_list_insert(attrs,
			pango_attr_background_new(0xffff, 0xffff, 0xffff));
	pango_layout_set_attributes(layout, attrs);

	sprintf(text, "%03u", 270);
	pango_layout_set_text(layout, text, strlen(text));
	pango_layout_get_pixel_size(layout, &tw, &th);
	tw += BG_TEXT_SIDE_SPACING;

	cx = ((double) width) / 2.0;
	cy = ((double) height) / 2.0;

	if (width < height)
		radius = ((width / 2 - tw) / 12) * 12;
	else
		radius = ((height / 2 - tw) / 12) * 12;
	step = radius / 6;

	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, width, height);
	if (NULL == pixbuf) {
		return -ENOMEM;
	}

	gdk_window_set_cursor(radar->window->window, radar->busy_cursor);
	gdk_display_sync(gdk_drawable_get_display(radar->window->window));

	radar_draw_bg_pixmap(radar, radar->canvas->window, pixbuf, 0xff,
			     layout, TRUE, cx, cy, step, radius, width, height);

	for (i = 0; i < RADAR_NR_VECTORS; i++) {
		v = &radar->vectors[i];

		if (!v->is_visible)
			continue;

		tvec.gc = v->gc;
		translate_point(v->x1, v->y1, radar->cx, radar->cy, radar->r,
				cx, cy, radius, &tvec.x1, &tvec.y1);
		translate_point(v->x2, v->y2, radar->cx, radar->cy, radar->r,
				cx, cy, radius, &tvec.x2, &tvec.y2);

		radar_draw_vector(radar, radar->canvas->window,
				  pixbuf, 0xff, TRUE, &tvec);
	}

	for (i = 0; i < RADAR_NR_SIGHTS; i++) {
		s = &radar->sights[i];

		for (j = 0; j < SIGHT_NR_VECTORS; j++) {
			v = &s->vectors[j];

			if (!v->is_visible)
				continue;

			tvec.gc = v->gc;
			translate_point(v->x1, v->y1,
					radar->cx, radar->cy, radar->r,
					cx, cy, radius, &tvec.x1, &tvec.y1);
			translate_point(v->x2, v->y2,
					radar->cx, radar->cy, radar->r,
					cx, cy, radius, &tvec.x2, &tvec.y2);

			radar_draw_vector(radar, radar->canvas->window,
					  pixbuf, 0xff, TRUE, &tvec);
		}

		for (j = 0; j < SIGHT_NR_ARCS; j++) {
			a = &s->arcs[j];

			if (!a->is_visible)
				continue;

			tarc.gc = a->gc;
			tarc.angle1 = a->angle1;
			tarc.angle2 = a->angle2;
			translate_point(a->x, a->y,
					radar->cx, radar->cy, radar->r,
					cx, cy, radius, &tarc.x, &tarc.y);
			translate_length(a->radius, radar->r,
					 radius, &tarc.radius);

			radar_draw_arc(radar, radar->canvas->window,
				       pixbuf, 0xff, TRUE, &tarc);
		}

		for (j = 0; j < SIGHT_NR_POLYS; j++) {
			p = &s->polys[j];

			if (!p->is_visible)
				continue;

			tpoly.gc = p->gc;
			tpoly.npoints = p->npoints;
			for (k = 0; k < tpoly.npoints; k++) {
				translate_point(p->points[k].x, p->points[k].y,
						radar->cx, radar->cy, radar->r,
						cx, cy, radius,
						&tpoly.points[k].x,
						&tpoly.points[k].y);
			}

			radar_draw_poly(radar, radar->canvas->window,
					pixbuf, 0xff, TRUE, &tpoly);
		}
	}


	gerror = NULL;
	if (gdk_pixbuf_savev(pixbuf, filename, type, keys, values, &gerror)) {
		error = -EIO;
		goto out;
	}

out:
	gdk_window_set_cursor(radar->window->window, NULL);

	g_object_unref(pixbuf);
	g_object_unref(layout);
	pango_font_description_free(font_description);
	pango_attr_list_unref(attrs);
	return error;
}

/*
 * Print Radarplot:
 *
 * Plot Background and Vectors as Postscript,
 * Plot Calculated Values as Postscript.
 */

typedef struct {
	GtkWidget	*table;
	GtkTooltips	*tooltips;
	int		rows_used;
	int		cols_used;
	GtkWidget	*width_label;
	GtkSpinButton	*width_spin;
	GtkWidget	*height_label;
	GtkSpinButton	*height_spin;
	GtkToggleButton	*aspect_toggle;
	int		width;
	int		height;
	gboolean	keep_aspect;
	double		aspect;
	int		change_level;
} size_chooser_t;

static void
sizer_width_changed(GtkSpinButton *spin, gpointer user_data)
{
	size_chooser_t *sizer = user_data;

	sizer->change_level++;

	sizer->width = gtk_spin_button_get_value_as_int(spin);
	if (sizer->keep_aspect && (sizer->change_level == 1)) {
		sizer->height = ((double) sizer->width) / sizer->aspect;
		gtk_spin_button_set_value(sizer->height_spin, sizer->height);
	}

	sizer->change_level--;
}

static void
sizer_height_changed(GtkSpinButton *spin, gpointer user_data)
{
	size_chooser_t *sizer = user_data;

	sizer->change_level++;

	sizer->height = gtk_spin_button_get_value_as_int(spin);
	if (sizer->keep_aspect && (sizer->change_level == 1)) {
		sizer->width = ((double) sizer->height) * sizer->aspect;
		gtk_spin_button_set_value(sizer->width_spin, sizer->width);
	}

	sizer->change_level--;
}

static void
sizer_aspect_toggled(GtkToggleButton *toggle, gpointer user_data)
{
	size_chooser_t *sizer = user_data;

	if (gtk_toggle_button_get_active(toggle)) {
		sizer->keep_aspect = TRUE;
		sizer->aspect = ((double) sizer->width) /
				((double) sizer->height);
	} else {
		sizer->keep_aspect = FALSE;
	}
}

static void
radar_init_size_chooser(size_chooser_t *sizer, int width, int height,
			int extra_rows, int extra_cols)
{
	memset(sizer, 0, sizeof(size_chooser_t));

	sizer->table = gtk_table_new(2 + extra_rows, 4 + extra_cols, FALSE);
	sizer->tooltips = gtk_tooltips_new();

	gtk_container_set_border_width(GTK_CONTAINER(sizer->table), 5);
	gtk_table_set_row_spacings(GTK_TABLE(sizer->table), TABLE_ROW_SPACING);
	gtk_table_set_col_spacings(GTK_TABLE(sizer->table), TABLE_COL_SPACING);

	sizer->width_label = gtk_label_new(UTF8("Breite:"));
	gtk_misc_set_alignment(GTK_MISC(sizer->width_label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(sizer->table), sizer->width_label,
			 0, 1, 0, 1,
			 0, GTK_EXPAND, 0, 0);
	sizer->width_spin = GTK_SPIN_BUTTON(
				radar_init_spin(320, 2048, 1, 10, width));
	gtk_tooltips_set_tip(sizer->tooltips, GTK_WIDGET(sizer->width_spin),
			     UTF8("Breite der Grafik in Pixeln"), NULL);
	g_signal_connect(G_OBJECT(sizer->width_spin), "value-changed",
			 G_CALLBACK(sizer_width_changed), sizer);
	gtk_table_attach(GTK_TABLE(sizer->table),
			 GTK_WIDGET(sizer->width_spin),
			 1, 2, 0, 1,
			 0, GTK_EXPAND, 0, 0);

	sizer->height_label = gtk_label_new(UTF8("Hhe:"));
	gtk_misc_set_alignment(GTK_MISC(sizer->height_label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(sizer->table), sizer->height_label,
			 2, 3, 0, 1,
			 0, GTK_EXPAND, 0, 0);
	sizer->height_spin = GTK_SPIN_BUTTON(
				radar_init_spin(320, 2048, 1, 10, height));
	gtk_tooltips_set_tip(sizer->tooltips, GTK_WIDGET(sizer->height_spin),
			     UTF8("Hhe der Grafik in Pixeln"), NULL);
	g_signal_connect(G_OBJECT(sizer->height_spin), "value-changed",
			 G_CALLBACK(sizer_height_changed), sizer);
	gtk_table_attach(GTK_TABLE(sizer->table),
			 GTK_WIDGET(sizer->height_spin),
			 3, 4, 0, 1,
			 0, GTK_EXPAND, 0, 0);

	sizer->aspect_toggle =
			GTK_TOGGLE_BUTTON(gtk_check_button_new_with_label(
				UTF8("Seitenverhltnis beibehalten")));
	gtk_tooltips_set_tip(sizer->tooltips, GTK_WIDGET(sizer->aspect_toggle),
			     UTF8("Seitenverhltnis von Breite zu Hhe der Grafik beibehalten"), NULL);
	gtk_toggle_button_set_active(sizer->aspect_toggle, TRUE);
	g_signal_connect(G_OBJECT(sizer->aspect_toggle), "toggled",
			 G_CALLBACK(sizer_aspect_toggled), sizer);
	gtk_table_attach(GTK_TABLE(sizer->table),
			 GTK_WIDGET(sizer->aspect_toggle),
			 0, 4, 1, 2,
			 0, GTK_EXPAND, 0, 0);

	sizer->rows_used = 2;
	sizer->cols_used = 4;

	sizer->width = width;
	sizer->height = height;

	sizer->keep_aspect = TRUE;
	sizer->aspect = ((double) width) / ((double) height);
}

void
radar_export_jpeg(GtkAction *action, gpointer user_data)
{
	char quality_as_string[32];
	radar_t *radar = user_data;
	size_chooser_t sizer;
	GtkSpinButton *quality_spin;
	GtkWidget *quality_label;
	GtkFileChooser *chooser;
	GtkFileFilter *filter;
	GtkWidget *dialog;
	char *p, *filename;
	char *pathname;
	char *options[2];
	char *keys[2];

	dialog = gtk_file_chooser_dialog_new(UTF8("JPEG Grafik speichern"),
					     GTK_WINDOW(radar->window),
					     GTK_FILE_CHOOSER_ACTION_SAVE,
					     GTK_STOCK_CANCEL,
					     GTK_RESPONSE_CANCEL,
					     GTK_STOCK_SAVE,
					     GTK_RESPONSE_ACCEPT,
					     NULL);
	chooser = GTK_FILE_CHOOSER(dialog);

	gtk_file_chooser_set_local_only(chooser, TRUE);
	gtk_file_chooser_set_select_multiple(chooser, FALSE);

	filter = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filter, "*.jpg");
	gtk_file_filter_add_pattern(filter, "*.jpeg");
	gtk_file_filter_set_name(filter, "JPEG Grafiken");
	gtk_file_chooser_set_filter(chooser, filter);

	if (radar->image_pathname) {
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
						    radar->image_pathname);
	}

	radar_init_size_chooser(&sizer, radar->w, radar->h, 0, 2);

	quality_label = gtk_label_new(UTF8("JPEG Qualitt:"));
	gtk_misc_set_alignment(GTK_MISC(quality_label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(sizer.table), quality_label,
			 sizer.cols_used, sizer.cols_used + 1, 0, 1,
			 0, GTK_EXPAND, 0, 0);
	quality_spin = GTK_SPIN_BUTTON(radar_init_spin(0, 100, 1, 10, 95));
	gtk_tooltips_set_tip(sizer.tooltips, GTK_WIDGET(quality_spin),
			     UTF8("JPEG Qualitt in Prozent (0 - 100)"), NULL);
	gtk_table_attach(GTK_TABLE(sizer.table), GTK_WIDGET(quality_spin),
			 sizer.cols_used + 1, sizer.cols_used + 2, 0, 1,
			 0, GTK_EXPAND, 0, 0);
	gtk_widget_show_all(sizer.table);

	gtk_file_chooser_set_extra_widget(chooser, sizer.table);

	if (gtk_dialog_run(GTK_DIALOG(dialog)) != GTK_RESPONSE_ACCEPT)
		goto out;

	pathname = gtk_file_chooser_get_current_folder(
						GTK_FILE_CHOOSER(dialog));
	if (pathname) {
		if (radar->image_pathname) {
			g_free(radar->image_pathname);
		}
		radar->image_pathname = pathname;
	}

	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));

	p = strrchr(filename, '.');
	if ((NULL == p) || (strcasecmp(p, ".jpg") && strcasecmp(p, ".jpeg"))) {
		p = g_malloc(strlen(filename) + 5);
		sprintf(p, "%s.jpg", filename);
		g_free(filename);
		filename = p;
	}

	sprintf(quality_as_string, "%u",
		gtk_spin_button_get_value_as_int(quality_spin));

	keys[0] = "quality";
	keys[1] = NULL;
	options[0] = quality_as_string;
	options[1] = NULL;

	gdk_window_set_cursor(dialog->window, radar->busy_cursor);

	radar_save_as_image(radar, sizer.width, sizer.height,
			    filename, "jpeg", keys, options);

	g_free(filename);

out:
	gtk_widget_destroy(dialog);
}

void
radar_export_png(GtkAction *action, gpointer user_data)
{
	radar_t *radar = user_data;
	size_chooser_t sizer;
	GtkFileChooser *chooser;
	GtkFileFilter *filter;
	GtkWidget *dialog;
	char *p, *filename;
	char *pathname;
	char *options[1];
	char *keys[1];

	dialog = gtk_file_chooser_dialog_new(UTF8("PNG Grafik speichern"),
					     GTK_WINDOW(radar->window),
					     GTK_FILE_CHOOSER_ACTION_SAVE,
					     GTK_STOCK_CANCEL,
					     GTK_RESPONSE_CANCEL,
					     GTK_STOCK_SAVE,
					     GTK_RESPONSE_ACCEPT,
					     NULL);
	chooser = GTK_FILE_CHOOSER(dialog);

	gtk_file_chooser_set_local_only(chooser, TRUE);
	gtk_file_chooser_set_select_multiple(chooser, FALSE);

	filter = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filter, "*.png");
	gtk_file_filter_set_name(filter, "PNG Grafiken");
	gtk_file_chooser_set_filter(chooser, filter);

	if (radar->image_pathname) {
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
						    radar->image_pathname);
	}

	radar_init_size_chooser(&sizer, radar->w, radar->h, 0, 0);
	gtk_widget_show_all(sizer.table);

	gtk_file_chooser_set_extra_widget(chooser, sizer.table);

	if (gtk_dialog_run(GTK_DIALOG(dialog)) != GTK_RESPONSE_ACCEPT)
		goto out;

	pathname = gtk_file_chooser_get_current_folder(
						GTK_FILE_CHOOSER(dialog));
	if (pathname) {
		if (radar->image_pathname) {
			g_free(radar->image_pathname);
		}
		radar->image_pathname = pathname;
	}

	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));

	p = strrchr(filename, '.');
	if ((NULL == p) || strcasecmp(p, ".png")) {
		p = g_malloc(strlen(filename) + 5);
		sprintf(p, "%s.png", filename);
		g_free(filename);
		filename = p;
	}

	keys[0] = NULL;
	options[0] = NULL;

	gdk_window_set_cursor(dialog->window, radar->busy_cursor);

	radar_save_as_image(radar, sizer.width, sizer.height,
			    filename, "png", keys, options);

	g_free(filename);

out:
	gtk_widget_destroy(dialog);
}
